/******************************************************************************
 *                                                                            *
 *                            Pikvorkov robot                               *
 *                                                                            *
 ******************************************************************************
 *
 * Autor:
 *   Josef Kufner <jkufner@seznam.cz>
 *
 *
 *
 * 
 * Popis Hrac plochy:
 * ~~~~~~~~~~~~~~~~~~~
 * - Hrac plocha je slovna od 0
 * - Jako prvn souadnice je vdy uvdn dek a pak teprve sloupec
 * - Hrac plocha je tvercov (poet dk a sloupc je stejn)
 * - Promnn 'size' udv velikost jedn strany hrac plochy
 *
 * Popis zpsobu komunikace:
 * ~~~~~~~~~~~~~~~~~~~~~~~~~
 * - Pokud je robot sputn s parametrem "-b", tak okamit po sputn
 *   vyle na stdout svj prvn tah.
 * - Pokud je robot sputn bez parametru "-b", tak vyk na protivnikv tah
 * 
 * - Vyslani tahu spov v zpisu etzce na stdout (viz. Send_Turn)
 * - Pjem tahu spov v peten etzce ze stdin (viz. Recv_Turn)
 * - Veker chybov hlen jsou zapisovna na stderr
 * 
 */

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#define 	DEBUG_PRINT		2	/* 2 = volitelne
						   1 = on
						   0 = off */

/******************************************************************************
 * DBG_MSG(str, arg...)  - Makro pro ladici vypisy
 */
#define		DBG_MSG_ALLWAYS(str, args...)				\
{					 				\
	fprintf(stderr, "Piskv-Rbt '%s': %s(%d): %d:", 			\
			name, __FILE__, __LINE__, dbg_level);		\
	for (dbg_i = 0; dbg_i < dbg_level; dbg_i++)			\
		fprintf(stderr, "   ");					\
	fprintf(stderr, str "\n", ##args);				\
	fflush(stderr);							\
}
#if DEBUG_PRINT == 2
#define		DBG_MSG(str, args...)					\
{									\
	if (debug_print) {						\
		DBG_MSG_ALLWAYS(str, ##args);				\
	}								\
}
#endif
#if DEBUG_PRINT == 1
#define		DBG_MSG(str, args...)	DBG_MSG_ALLWAYS(str, ##args)
#endif
#if DEBUG_PRINT == 0
#define		DBG_MSG(str, args...)
#endif

#define		DBG_FN_IN	{dbg_level++;}
#define		DBG_FN_OUT	{dbg_level--;}
/*****************************************************************************/

/* Makro pro jednodusi pristup k 1 rozmernemu poli jako k 2 rozmernemu 
 * x .. radek
 * y .. sloupec
 */ 
#define		PLAY_AREA(x, y) 	(play_area[(x) * size + (y)])
#define		HODNOCENI(x, y)		(hodnoceni[(x) * size + (y)])

/* Hodnoty poli na hraci ploce */
#define		P_NONE			0	/* Pole je volne */
#define		P_JA			1	/* Moje znacka */
#define		P_ON			2	/* Protihracova znacka */

/* Algoritmy */
#define		ALG_RANDOM		0
#define		ALG_SINGLE		1
#define		ALG_BACKTRACKING	2


#define		ROW			0
#define		COL			1

/*****************************************************************************/
int dbg_level = 0, dbg_i;

struct hodnoceni_line {
	int lenght;			/* Delka rady (1 .. 5) */
	int free_ends;			/* Pocet volnych koncu rady (0 .. 2) */
	int is_space;			/* volne misto na radu*/
};

struct hodnoceni_poli {
	int is_free;			/* Je pole volne */
	int row, col;			/* Umisteni policka */
	
	struct hodnoceni_line line[2][4];/* Rady vznikle tahem */
	long int body;			/* Bodove ohodnoceni */
	
} *hodnoceni;				/* Pole na ohodnoceni plochy */


int size = 15;
int *play_area;				/* Pole pro hraci plochu */
int tahy;				/* Pocet moznych tahu */

int first_turn = 0; 			/* Zacinam ? */ 
int alg = ALG_BACKTRACKING;		/* Hrat nahodne ? */
int debug_print = 0;			/* Ladici vypisy */

char name[256] = "";

const signed short int smer[8][2] = {	/* Pomocne pole pro scitani poli */
	/* row,	col */
	{   0,	 1  },	/* Horizontal */
	{   1,	 0  },	/* Vertical */
	{   1,	 1  },	/* Left (\) */
	{   1,	-1  },	/* Right (/) */
	{   0,	-1  },
	{  -1,	 0  },
	{  -1, 	-1  },
	{  -1,	 1  }
};

/* Bodove hodnoceni moznych tahu */
const long int body[/* lenght */][3 /* free_ends */] = {
	/* free_ends:
	 *  	0,		1,	 	2 */		   /* lenght */
	{	0,		0,	 	0 		},	/* 0 */
	{	1,		1,	 	1 		},	/* 1 */
	{	1,		10,	 	100 		},	/* 2 */
	{	1,		1000,	 	10000 		},	/* 3 */
	{	1,		100000,	 	1000000 	},	/* 4 */
	{	10000000,	10000000,	10000000	}	/* 5 */
};
const long int body_space = 3;


void On_Exit (void);
void On_Signal (int sig);
void Send_Turn(int row, int col);
void Recv_Turn (void);
int Turn_Comp (const void *item1, const void *item2);
void Random_Turn (int *turn_row, int *turn_col);
int Gen_Turn(int *turn_row, int *turn_col, int player);
void First_Turn (void);
int Turn (int *turn_row, int *turn_col, int loop_max);
int is_Game_Over(void);

/******************************************************************************
 * 	On_Exit - funkce volana pri ukonceni programu
 *
 */
void On_Exit (void) {
	DBG_MSG("On_Exit:");
	DBG_FN_IN;

	/* Uvolnovani pameti */
	if (play_area)
		free(play_area);
	if (hodnoceni)
		free(hodnoceni);

	DBG_FN_OUT;
}
/******************************************************************************
 * 	On_Signal - obsluha signalu
 */
void On_Signal (int sig) {
	switch (sig) {
		case SIGSEGV:
			DBG_MSG("SIGSEGV");
			DBG_FN_IN;
			exit(10);
			break;
		case SIGTERM:
			DBG_MSG("SIGTERM");
			DBG_FN_IN;
			exit(0);
	}
	DBG_FN_OUT;
}


/******************************************************************************
 * 	Send_Turn - Odeslani tahu
 *
 */
void Send_Turn(int row, int col) {
	DBG_MSG("Send_Turn:");
	DBG_FN_IN;
	
	DBG_MSG("Posilam: row = %d, col = %d...", row, col);
	
	printf("> row = %d, col = %d;\n", row, col);
	fflush(stdout);
	PLAY_AREA(row, col) = P_JA;
		
	DBG_MSG("Poslano.");

	DBG_FN_OUT;
	return;
}


/******************************************************************************
 * 	Recv_Turn - Prijem tahu
 *
 * Vycka na tah na stdin.
 *
 */
void Recv_Turn (void) {
 int row, col;
	
	DBG_MSG("Recv_Turn:");
	DBG_FN_IN;
	
	/* Nacteni tahu */
	DBG_MSG("Cekam na data...");
	if (getchar() == 0) {
		exit(0);
	}
	if (scanf(" row = %d, col = %d;", &row, &col) != 2) {
		fprintf(stderr, "Piskv-Rbt %s: Chyba pri nacitani tahu!\n", name);
		DBG_FN_OUT;
		exit(2);
	}
	getc(stdin);

	DBG_MSG("Prijato: row = %d, col = %d;", row, col);
	
	/* Kontrola tahu*/
	if (row < 0 || row >= size || col < 0 || col >= size) {
		fprintf(stderr, "Piskv-Rbt %s: Tah [%d, %d] je mimo plochu!\n", name, row, col);
		DBG_FN_OUT;
		exit(3);
	}
	if (PLAY_AREA(row, col) != P_NONE) {
		fprintf(stderr, "Piskv-Rbt %s: Tah [%d, %d] je na jiz obsazene pole!\n", name, row, col);
		DBG_FN_OUT;
		exit(3);
	}

	PLAY_AREA(row, col) = P_ON;
	DBG_FN_OUT;
	return ;
}

/******************************************************************************
 * 	Turn_Comp - Porovnavani tahu
 */
int Turn_Comp (const void *item1, const void *item2) {
	DBG_FN_IN;
	if ((*((struct hodnoceni_poli *)item1)).body > (*((struct hodnoceni_poli *)item2)).body) {
		DBG_FN_OUT;
		return -1;
	}

	if ((*((struct hodnoceni_poli *)item1)).body == (*((struct hodnoceni_poli *)item2)).body) {
		DBG_FN_OUT;
		return 0;
	}
	
	if ((*((struct hodnoceni_poli *)item1)).body < (*((struct hodnoceni_poli *)item2)).body) {
		DBG_FN_OUT;
		return 1;
	}
}


/******************************************************************************
 * 	Random_Turn - Generovani nahodneho tahu
 */
void Random_Turn (int *turn_row, int *turn_col) {
	DBG_MSG("Random_Turn:");
	DBG_FN_IN;
	
	*turn_row = *turn_col = size / 2;
	
	while (PLAY_AREA(*turn_row, *turn_col) != P_NONE) {
		*turn_row = rand() % size;
		*turn_col = rand() % size;
	}
		
	DBG_FN_OUT;
	return;
}


/******************************************************************************
 * 	Gen_Turn  -  Vytvoreni tahu vcetne propoctu dopredu
 */
int Gen_Turn(int *turn_row, int *turn_col, int player) {
 int row, col, max = -1, ret;
	
	do {
		max = Turn(&row, &col, max);
		PLAY_AREA(row, col) = player;
		switch (is_Game_Over()) {
			case P_JA:	/* OK, Vyhral jsem. Tyhle tahy plati. */
				*turn_row = row;
				*turn_col = col;
				PLAY_AREA(row, col) = P_NONE;
				return 1;
			case P_ON:	/* Tudy cesta nevede */
				PLAY_AREA(row, col) = P_NONE;
				return 0;
			case P_NONE:	/* Jeste nikdo nevyhral. Dalsi tahy. */
				player = (player == P_JA) ? P_ON : P_JA;
				if (Gen_Turn(turn_row, turn_col, player) == 1) {
					ret = 1;
				} else {
					ret = 0;
				}
				PLAY_AREA(row, col) = P_NONE;
				return ret;
		}
	} while (max < 0);
	return 0;
}


/******************************************************************************
 *	Turn - Vygenerovani tahu
 *
 */
int Turn (int *turn_row, int *turn_col, int loop_max) {
 int row, col, n, s, player, b, i, max;

 	DBG_MSG("Turn:");
	DBG_FN_IN;
	
	DBG_MSG("Analyza...");
	for (row = 0; row < size; row++) {
		for (col = 0; col < size; col++) {
			/* Zapis polohy tahu */
			HODNOCENI(row, col).row = row;
			HODNOCENI(row, col).col = col;
			
			/* Je pole volne ? */
			if (PLAY_AREA(row, col) != P_NONE) {
				HODNOCENI(row, col).is_free = 0;
				HODNOCENI(row, col).body = 0;
				continue;
			}
			HODNOCENI(row, col).is_free = 1;

			for (player = 1; player <= 2; player++) {
				for (s = 0; s < 8; s++) {
				
					/* Spocitani delek rad a volnych koncu */
					HODNOCENI(row, col).line[player][s % 4].free_ends = 0;
					n = 1;
					while (row + n * smer[s][ROW] < size && row + n * smer[s][ROW] >= 0 &&
							col + n * smer[s][COL] < size && col + n * smer[s][COL] >= 0) {
						if (PLAY_AREA(row + n * smer[s][ROW], col + n * smer[s][COL]) == player) {
							n++;
						} else {
							HODNOCENI(row, col).line[player][s % 4].free_ends++;
							break;
						}
					}
					if (s < 4)
						HODNOCENI(row, col).line[player][s % 4].lenght = n;
					else
						HODNOCENI(row, col).line[player][s % 4].lenght += n;

					/* Spocitani volneho mista */
					HODNOCENI(row, col).line[player][s % 4].is_space = 0;
					
					n = 1;
					for (i = 1; i <= 5 && row + i * smer[s][ROW] < size 
							&& row + i * smer[s][ROW] >= 0
							&& col + i * smer[s][COL] < size 
							&& col + i * smer[s][COL] >= 0
							&& ( PLAY_AREA(row + i * smer[s][ROW], 
								col + i * smer[s][COL]) == player ||
							PLAY_AREA(row + i * smer[s][ROW], 
								col + i * smer[s][COL]) == P_NONE);
							i++) {
						n++;
					}
					
					if (s < 4)
						HODNOCENI(row, col).line[player][s % 4].is_space = n;
					if (s >= 4 && HODNOCENI(row, col).line[player][s % 4].is_space + n >= 5)
						HODNOCENI(row, col).line[player][s % 4].is_space = 1;

				}
			}

			/* Bodove ohodnoceni moznych tahu */
			b = 0;
			for (player = 1; player <= 2; player++) {
				for (s = 0; s < 4; s++) {
					b += body
						[HODNOCENI(row, col).line[player][s].lenght]
						[HODNOCENI(row, col).line[player][s].free_ends];
					if (HODNOCENI(row, col).line[player][s].is_space == 0) 
						b /= 2;	
				}
			}
			if (HODNOCENI(row, col).is_free)
				HODNOCENI(row, col).body = b;
			else
				HODNOCENI(row, col).body = 0;

			
		}
	}
	
	/* Vyber tahu */
	DBG_MSG("Vyber tahu...");
	for (i = 0, max = 0, n = 0; i < size * size; i++) {
		if (max < hodnoceni[i].body && (loop_max == -1 || hodnoceni[i].body <= loop_max)) {
			n = i;
			max = hodnoceni[i].body;
		}
	}
	
	/* Vraceni tahu */
	*turn_row = hodnoceni[n].row;
	*turn_col = hodnoceni[n].col;
	DBG_MSG("Tah: row = %d, col = %d", *turn_row, *turn_col);
	DBG_FN_OUT;
	return max;
}


/******************************************************************************
 * 	First_Turn - Provedeni prvniho tahu
 *
 * Tahne na prostredni pole
 *
 */
void First_Turn (void) {
	DBG_MSG("First_Turn:");
	DBG_FN_IN;
	Send_Turn(size / 2, size / 2);
	DBG_FN_OUT;
	return;
}


/******************************************************************************
 * 	is_Game_Over - Kontrola konce hry
 *
 */
int is_Game_Over(void) {
	DBG_MSG("is_Game_Over:");
	DBG_FN_IN;
	DBG_MSG("Nekontroluji a vracim 0.");
	DBG_FN_OUT;
	return 0;
}


/******************************************************************************
 * 	main
 */
int main (int argc, char *argv[]) {
 int opt, row, col;

	DBG_MSG("Start...");
	DBG_FN_IN;
	DBG_MSG("main:");
 	
	/* Nastaveni obsluhy signalu */
	signal(SIGSEGV, On_Signal);
	signal(SIGTERM, On_Signal);
	
	/* Inicializace generatoru nahodnych cisel */
 	srand((unsigned)time(NULL));
 
 	/* Dekodovani parametru  */
	while ((opt = getopt(argc, argv, "a:n:bs:rhd")) != -1) {
		switch(opt) {
		case 'n': /* Jmeno hrace */
			strcpy(name, optarg);
			break;
		case 'r': /* Nahodne tahy */
			alg = ALG_RANDOM;
			break;
		case 'a': /* Algoritmus */
			switch (optarg[0]) {
			case '0':
				alg = ALG_RANDOM;
				break;
			case '1':
				alg = ALG_SINGLE;
				break;
			case '2':
				alg = ALG_BACKTRACKING;
				break;
			default:
				fprintf(stderr, "Piskv-Rbt: Chybny parametr prepinace -a\n");
			}
			break;
		case 'b': /* Zacinam ? */
			first_turn = 1;
			break;
		case 's': /* Nastaveni velikosti hraci plochy  */
			if (sscanf(optarg, "%d", &size) == 1) {
				if (size < 5)
					size = 15;
			} else {
				size = 15;
			}
			break;
		case 'd': /* Ladici vypisy */
			debug_print = 1;
			break;
		case '?': /* Napoveda */
		case 'h':
			fprintf(stderr, "Pouziti: piskv-rbt [-h] [-b] [-s size] [-r] [-n name] [-d]\n"
					"\t-h ... Vypise tuto napovedu\n"
					"\t-b ... Pokyn k prvnimu tahu\n"
					"\t-s ... Nastaveni velikosti hraci plochy\n"
					"\t-r ... Generovat nahodne tahy\n"
					"\t-n ... Pojmenovani hrace (debug)\n"
					"\t-d ... Debug (volitelne)\n"
					"\n"
					"Tento program lze it a modifikovat v souladu\n"
					"s podmnkami licence GNU Public Licence.\n"
					"Pouvn tohoto programu je BEZ JAKKOLIV ZRUKY\n"
					"a vhradn na VLASTN NEBEZPE.\n"
					"\n");
			DBG_FN_OUT;
			return 1;
			break;
		}
	}

	/* Alokace hraci plochy */
	play_area = (int *) malloc(sizeof(int) * size * size);
	if (play_area == NULL) {
		fprintf(stderr, "Piskv-Rbt: Chyba pri alokaci pameti (%d): %s", errno, strerror(errno));
		DBG_FN_OUT;
		return 2;
	}
	
	/* Zaregistrovani uklizeci funkce */
	atexit(On_Exit);
	
	/* Alokace hodnoticiho pole struktur */
	hodnoceni = (struct hodnoceni_poli *) malloc(sizeof(struct hodnoceni_poli) * size * size);
	if (hodnoceni == NULL) {
		fprintf(stderr, "Piskv-Rbt: Chyba pri alokaci pameti (%d): %s", errno, strerror(errno));
		DBG_FN_OUT;
		return 2;
	}

	/* Vycisteni hraci plochy */
	for (row = 0; row < size; row++) {
		for (col = 0; col < size; col++) {
			PLAY_AREA(row, col) = P_NONE;
		}
	}

	/* Hra ! */
	if (first_turn) {
		First_Turn();
	}
	while(!is_Game_Over()) {

		Recv_Turn();
		
		if(is_Game_Over())
			break;
		
		switch (alg) {
			case ALG_RANDOM:
				Random_Turn(&row, &col);
				break;
			case ALG_SINGLE:
				Turn(&row, &col, -1);
				break;
			default:
			case ALG_BACKTRACKING:
				Gen_Turn(&row, &col, P_JA);
				break;
		}
		
		Send_Turn(row, col);
	}
	
	DBG_FN_OUT;
	return 0;
}

